from deta_discord_interactions.utils.database.database import Database
from deta_discord_interactions.utils.database.record import Record, AutoSyncRecord
from deta_discord_interactions.utils.database.query import Query, Field
from deta_discord_interactions.utils.database.record import RecordType, Key

__all__ = [
    "Database",
    "Record",
    "AutoSyncRecord",
    "Query",
    "Field",
    "RecordType",
    "Key",
]



import os
import time
from typing import Callable, Dict, List, NoReturn
import json

import requests

from nacl.exceptions import BadSignatureError
from nacl.signing import VerifyKey

from deta_discord_interactions.models.option import Option

from deta_discord_interactions.command import Command, SlashCommandGroup
from deta_discord_interactions.context import Context, ApplicationCommandType
from deta_discord_interactions.models import Message, Modal, ResponseType, Permission


class AbortError(Exception):
    def __init__(self, http_code):
        self.http_code = http_code

class PongResponse:
    def encode(self):
        return json.dumps({"type": ResponseType.PONG}), "application/json"


class InteractionType:
    PING = 1
    APPLICATION_COMMAND = 2
    MESSAGE_COMPONENT = 3
    APPLICATION_COMMAND_AUTOCOMPLETE = 4
    MODAL_SUBMIT = 5


class DiscordInteractionsBlueprint:
    """
    Represents a collection of :class:`ApplicationCommand` s.

    Useful for splitting a bot across multiple files.
    """

    def __init__(self):
        self.discord_commands = {}
        self.custom_id_handlers = {}
        self.tasks = []

    def add_command(
        self,
        command: Callable,
        name: str = None,
        description: str = None,
        *,
        options: List[Option] = None,
        annotations: Dict[str, str] = None,
        type: int = ApplicationCommandType.CHAT_INPUT,
        default_member_permissions: int = None,
        dm_permission: bool = None,
        name_localizations: Dict[str, str] = None,
        description_localizations: Dict[str, str] = None,
    ):
        """
        Create and add a new :class:`ApplicationCommand`.

        Parameters
        ----------
        command: Callable
            Function to execute when the command is run.
        name: str
            The name of the command, as displayed in the Discord client.
        name_localizations: Dict[str, str]
            A dictionary of localizations for the name of the command.
        description: str
            The description of the command.
        description_localizations: Dict[str, str]
            A dictionary of localizations for the description of the command.
        options: List[Option]
            A list of options for the command, overriding the function's
            keyword arguments.
        annotations: Dict[str, str]
            If ``options`` is not provided, descriptions for each of the
            options defined in the function's keyword arguments.
        type: int
            The class:`.ApplicationCommandType` of the command.
        default_member_permissions: int
            A permission integer defining the required permissions a user must have to run the command.
        dm_permission: bool
            Indicates whether the command can be used in DMs.
        """
        command = Command(
            command=command,
            name=name,
            description=description,
            options=options,
            annotations=annotations,
            type=type,
            default_member_permissions=default_member_permissions,
            dm_permission=dm_permission,
            name_localizations=name_localizations,
            description_localizations=description_localizations,
            discord=self,
        )
        self.discord_commands[command.name] = command
        return command

    def command(
        self,
        name: str = None,
        description: str = None,
        *,
        options: List[Option] = None,
        annotations: Dict[str, str] = None,
        type: int = ApplicationCommandType.CHAT_INPUT,
        default_member_permissions: int = None,
        dm_permission: bool = None,
        name_localizations: Dict[str, str] = None,
        description_localizations: Dict[str, str] = None,
    ) -> Callable[[Callable], Command]:
        """
        Decorator to create a new :class:`Command`.

        Parameters
        ----------
        name: str
            The name of the command, as displayed in the Discord client.
        name_localizations: Dict[str, str]
            A dictionary of localizations for the name of the command.
        description: str
            The description of the command.
        description_localizations: Dict[str, str]
            A dictionary of localizations for the description of the command.
        options: List[Option]
            A list of options for the command, overriding the function's
            keyword arguments.
        annotations: Dict[str, str]
            If ``options`` is not provided, descriptions for each of the
            options defined in the function's keyword arguments.
        type: int
            The ``ApplicationCommandType`` of the command.
        default_member_permissions: int
            A permission integer defining the required permissions a user must have to run the command
        dm_permission: bool
            Indicates whether the command can be used in DMs

        Returns
        -------
        Callable[Callable, Command]
        """

        def decorator(func):
            nonlocal name, description, type, options
            command = self.add_command(
                func,
                name=name,
                description=description,
                options=options,
                annotations=annotations,
                type=type,
                default_member_permissions=default_member_permissions,
                dm_permission=dm_permission,
                name_localizations=name_localizations,
                description_localizations=description_localizations,
            )
            return command

        return decorator

    def task(self):
        """
        Decorator to add a new Task.
        The Micro must be set to run on CRON for them to run at all.
        """
        def decorator(func):
            self.tasks.append(func)
            return func
        return decorator

    def command_group(
        self,
        name: str,
        description: str = "No description",
        *,
        default_member_permissions: int = None,
        dm_permission: bool = None,
        name_localizations: Dict[str, str] = None,
        description_localizations: Dict[str, str] = None,
    ):
        """
        Create a new :class:`SlashCommandGroup`
        (which can contain multiple subcommands)

        Parameters
        ----------
        name: str
            The name of the command group, as displayed in the Discord client.
        name_localizations: Dict[str, str]
            A dictionary of localizations for the name of the command group.
        description: str
            The description of the command group.
        description_localizations: Dict[str, str]
            A dictionary of localizations for the description of the command group.
        default_member_permissions: int
            A permission integer defining the required permissions a user must have to run the command
        dm_permission: bool
            Indicates whether the command can be used in DMs

        Returns
        -------
        SlashCommandGroup
            The newly created command group.
        """

        group = SlashCommandGroup(
            name=name,
            description=description,
            default_member_permissions=default_member_permissions,
            dm_permission=dm_permission,
            name_localizations=name_localizations,
            description_localizations=description_localizations,
        )
        self.discord_commands[name] = group
        return group

    def add_custom_handler(self, handler: Callable, custom_id: str):
        """
        Add a handler for an incoming interaction with the specified custom ID.

        Parameters
        ----------
        handler: Callable
            The function to call to handle the incoming interaction.
        custom_id: str
            The custom ID to respond to.

        Returns
        -------
        str
            The custom ID that the handler will respond to.
        """
        self.custom_id_handlers[custom_id] = handler
        return custom_id

    def custom_handler(self, custom_id: str):
        """
        Returns a decorator to register a handler for a custom ID.

        Parameters
        ----------
        custom_id
            The custom ID to respond to.

        Returns
        -------
        Callable[Callable, str]
        """

        def decorator(func):
            nonlocal custom_id
            custom_id = self.add_custom_handler(func, custom_id)
            return custom_id

        return decorator


class DiscordInteractions(DiscordInteractionsBlueprint):
    """
    Handles registering a collection of :class:`Command` s, receiving
    incoming interaction data, and sending/editing/deleting messages via
    webhook.
    """
    DISCORD_BASE_URL = "https://discord.com/api/v10"

    def __init__(self):
        super().__init__()
        self.discord_token = None
        try:
            self.discord_client_id = os.environ["DISCORD_CLIENT_ID"]
            self.discord_public_key = os.environ["DISCORD_PUBLIC_KEY"]
            self.discord_client_secret = os.environ["DISCORD_CLIENT_SECRET"]
            self.discord_scope = os.getenv("DISCORD_SCOPE", "applications.commands.update")
            self.DONT_REGISTER_WITH_DISCORD = os.getenv("DONT_REGISTER_WITH_DISCORD", False)
            self.DONT_VALIDATE_SIGNATURE = os.getenv("DONT_VALIDATE_SIGNATURE", False)
        except KeyError:
            raise Exception("Please fill in the .env files with your application's credentials.")
        self.__routes = {}
        # For Deta Micro CRON
        self._deta_type = "probably_wsgi"
        self.lib = self.run_tasks

    def fetch_token(self):
        """
        Fetch an OAuth2 token from Discord using the ``CLIENT_ID`` and
        ``CLIENT_SECRET`` with the ``applications.commands.update`` scope. This
        can be used to register new application commands.
        """
        if self.DONT_REGISTER_WITH_DISCORD:
            discord_token = {
                "token_type": "Bearer",
                "scope": self.discord_scope,
                "expires_in": 604800,
                "access_token": "DONT_REGISTER_WITH_DISCORD",
            }
            discord_token["expires_on"] = (
                time.time() + discord_token["expires_in"] / 2
            )
            return discord_token

        response = requests.post(
            self.DISCORD_BASE_URL + "/oauth2/token",
            data={
                "grant_type": "client_credentials",
                "scope": self.discord_scope,
            },
            headers={"Content-Type": "application/x-www-form-urlencoded"},
            auth=(self.discord_client_id, self.discord_client_secret),
        )

        response.raise_for_status()
        discord_token = response.json()
        discord_token["expires_on"] = (
            time.time() + discord_token["expires_in"] / 2
        )
        return discord_token

    def auth_headers(self):
        """
        Get the Authorization header required for HTTP requests to the
        Discord API.

        Returns
        -------
        Dict[str, str]
            The Authorization header.
        """

        if self.discord_token is None or time.time() > self.discord_token["expires_on"]:
            self.discord_token = self.fetch_token()
        return {"Authorization": f"Bearer {self.discord_token['access_token']}"}

    def update_commands(self, guild_id: str = None):
        """
        Update the list of commands registered with Discord.
        This method will overwrite all existing commands.

        Parameters
        ----------
        guild_id: str
            The ID of the Discord guild to register commands to. If omitted,
            the commands are registered globally.
        """
        if os.getenv("DETA_RUNTIME") is not None:  # It *would* work, it's just a big waste and may slow down the bot overall
            raise Exception("Cannot register commands from inside a Deta Micro")

        if guild_id:
            url = (
                f"{self.DISCORD_BASE_URL}/applications/"
                f"{self.discord_client_id}/"
                f"guilds/{guild_id}/commands"
            )
        else:
            url = (
                f"{self.DISCORD_BASE_URL}/applications/"
                f"{self.discord_client_id}/commands"
            )

        overwrite_data = [command.dump() for command in self.discord_commands.values()]

        if not self.DONT_REGISTER_WITH_DISCORD:
            response = requests.put(
                url, json=overwrite_data, headers=self.auth_headers()
            )

            try:
                response.raise_for_status()
            except requests.exceptions.HTTPError:
                raise ValueError(
                    f"Unable to register commands:"
                    f"{response.status_code} {response.text}"
                )

            for command in response.json():
                if command["name"] in self.discord_commands:
                    self.discord_commands[command["name"]].id = command["id"]
        else:
            for command in self.discord_commands.values():
                command.id = command.name

    def build_permission_overwrite_url(
        self,
        command: Command = None,
        *,
        guild_id: str,
        command_id: str = None,
    ):
        """
        Build the URL for getting or setting permission overwrites for a
        specific guild and command.
        """
        if command_id is None:
            if command:
                command_id = command.id
            else:
                raise ValueError(
                    "You must supply either a command ID or a Command instance."
                )

        url = f"{self.DISCORD_BASE_URL}/applications/{self.discord_client_id}/guilds/{guild_id}/commands/{command_id}/permissions"

        return url

    def get_permission_overwrites(
        self,
        command: Command = None,
        *,
        guild_id: str,
        command_id: str = None,
    ):
        """
        Get the list of permission overwrites in a specific guild for a
        specific command.

        Parameters
        ----------
        command: Command
            The :class:`.Command` to retrieve permissions for.
        guild_id: str
            The ID of the guild to retrieve permissions from.
        command_id: str
            The ID of the command to retrieve permissions for.

        Returns
        -------
        List[Permission]
            A list of permission overwrites for the given command.
        """

        url = self.build_permission_overwrite_url(
            command,
            guild_id=guild_id,
            command_id=command_id,
        )

        response = requests.get(
            url,
            headers=self.auth_headers(),
        )
        response.raise_for_status()

        return [Permission.from_dict(perm) for perm in response.json()]

    def set_permission_overwrites(
        self,
        permissions: List[Permission],
        command: Command = None,
        *,
        guild_id: str,
        command_id: str = None,
    ):
        """
        Overwrite the list of permission overwrites in a specific guild for a
        specific command.

        Parameters
        ----------
        command: Command
            The :class:`.Command` to retrieve permissions for.
        guild_id: str
            The ID of the guild to retrieve permissions from.
        command_id: str
            The ID of the command to retrieve permissions for.
        """

        url = self.build_permission_overwrite_url(
            command,
            guild_id=guild_id,
            command_id=command_id,
        )

        response = requests.put(
            url,
            headers=self.auth_headers(),
            json={"permissions": [perm.dump() for perm in permissions]},
        )
        response.raise_for_status()


    def register_blueprint(self, blueprint: DiscordInteractionsBlueprint):
        """
        Register a :class:`DiscordInteractionsBlueprint` to this
        DiscordInteractions class. Updates this instance's list of
        :class:`Command` s using the blueprint's list of
        :class:`Command` s.

        Parameters
        ----------
        blueprint: DiscordInteractionsBlueprint
            The :class:`DiscordInteractionsBlueprint` to add
            :class:`Command` s from.
        """
        self.discord_commands.update(blueprint.discord_commands)
        self.custom_id_handlers.update(blueprint.custom_id_handlers)
        self.tasks.extend(blueprint.tasks)

    def run_command(self, data: dict):
        """
        Run the corresponding :class:`Command` given incoming interaction
        data.

        Parameters
        ----------
        data
            Incoming interaction data.

        Returns
        -------
        Message
            The resulting message from the command.
        """

        command_name = data["data"]["name"]

        command = self.discord_commands.get(command_name)

        if command is None:
            raise ValueError(f"Invalid command name: {command_name}")

        return command.make_context_and_run(discord=self, data=data)

    def run_handler(self, data: dict, *, allow_modal: bool = True):
        """
        Run the corresponding custom ID handler given incoming interaction
        data.

        Parameters
        ----------
        data
            Incoming interaction data.

        Returns
        -------
        Message
            The resulting message.
        """

        context = Context.from_data(self, data)
        handler = self.custom_id_handlers[context.primary_id]
        args = context.create_handler_args(handler)
        result = handler(context, *args)

        if isinstance(result, Modal):
            if allow_modal:
                return result
            else:
                raise ValueError("Cannot return a Modal to that interaction type.")

        return Message.from_return_value(result)

    def run_autocomplete(self, data: dict):
        """
        Run the corresponding command's autocomplete handler given incoming 
        interaction data.

        Parameters
        ----------
        data
            Incoming interaction data.

        Returns
        -------
        AutocompleteResult
            The result of the autocomplete handler.
        """

        command_name = data["data"]["name"]

        command = self.discord_commands[command_name]

        return command.make_context_and_run_autocomplete(discord=self, data=data)

    def run_tasks(self, event):
        """Runs all registered Tasks. 
        If there are more than one registerd tasks, they may run asynchronously and out of order."""
        from concurrent.futures import ThreadPoolExecutor
        with ThreadPoolExecutor() as executor:
            for task in self.tasks:
                executor.submit(task)


    def verify_signature(self, request):
        """
        Verify the signature sent by Discord with incoming interactions.

        Parameters
        ----------
        request
            The request to verify the signature of.
        """
        signature = request.get("HTTP_X_SIGNATURE_ED25519")
        timestamp = request.get("HTTP_X_SIGNATURE_TIMESTAMP")

        if self.DONT_VALIDATE_SIGNATURE:
            return

        if signature is None or timestamp is None:
            self.abort(401, "Missing signature or timestamp")

        message = f"{timestamp}{request['raw_data'].decode('UTF-8')}".encode("UTF-8")
        verify_key = VerifyKey(bytes.fromhex(self.discord_public_key))
        try:
            verify_key.verify(message, bytes.fromhex(signature))
        except BadSignatureError:
            try:
                body = (
                    json.dumps(
                        json.loads(
                            request['raw_data'].decode('UTF-8')
                        ),
                        separators=(',', ':'),
                    )
                    .encode('UTF-8')
                )
                message = f"{timestamp}{body}".encode("UTF-8")
                verify_key.verify(message, bytes.fromhex(signature))
            except BadSignatureError:
                self.abort(401, "Incorrect Signature")
            else:
                import warnings
                warnings.warn("The whitespace for the request data may have been modified before being sent to discord-interactions")

    def handle_interaction(self, request: dict):
        """
        Verify the signature in the incoming request and return the
        result from the corresponding handler.

        Returns
        -------
        Message | Modal | AutocompleteResult
            The response from the corresponding handler.
        """
        self.verify_signature(request)

        interaction_type = request["json"].get("type")
        if interaction_type == InteractionType.PING:
            return PongResponse()
        elif interaction_type == InteractionType.APPLICATION_COMMAND:
            return self.run_command(request["json"])
        elif interaction_type == InteractionType.MESSAGE_COMPONENT:
            return self.run_handler(request["json"])
        elif interaction_type == InteractionType.APPLICATION_COMMAND_AUTOCOMPLETE:
            return self.run_autocomplete(request["json"])
        elif interaction_type == InteractionType.MODAL_SUBMIT:
            return self.run_handler(request["json"], allow_modal=False)
        else:
            raise RuntimeWarning(
                f"Interaction type {interaction_type} is not yet supported"
            )
    
    def route(self, route_path: str):
        """Decorator to register a custom route.
        Used internally for Webhooks OAuth, but for most purposes,
        I would recommend to use another Micro with FastAPI or Flask instead.

        Example usage:
        @app.route('/')
        def home(request, start_response, abort):
            start_response('200 OK', [])
            return ['Hello World!'.encode('UTF-8')]
        """
        def decorator(function):
            self.__routes[route_path] = function
            return function
        return decorator

    def __call__(self, environ, start_response):
        """
        Handles incoming interaction data
        (WSGI)
        """
        try:
            data = environ.copy()
            raw_data = environ["wsgi.input"].read()
            if raw_data:
                data["json"] = json.loads(raw_data.decode("UTF-8"))
                data["raw_data"] = raw_data

            data["path"] = data.get("PATH_INFO", '') or '/'

            if data["QUERY_STRING"]:
                data["query_dict"] = dict(args.split('=', 1) for args in data["QUERY_STRING"].split("&"))

            if data['path'] == '/discord':
                result = self.handle_interaction(data)
                response, mimetype = result.encode()
                status = "200 OK"
                response_headers = [("Content-Type", mimetype)]
                start_response(status, response_headers)
                return [response.encode("UTF-8")]
            elif (  # If you set it like `https://example.deta.dev` instead of `https://example.deta.dev/discord`
                data['path'] == '/' 
                and '/' not in self.__routes 
                and "Discord-Interactions" in data.get("HTTP_USER_AGENT")
                and self.verify_signature(data)
            ):
                raise Exception("Please set the path to `.../discord` on the Developer Portal, not just the Micro URL")
            elif data['path'] not in self.__routes:
                self.abort(404, 'Page not found')
            else:
                return self.__routes[data['path']](data, start_response, self.abort)
        except AbortError as err:
            status = err.http_code
            response_headers = [("Content-Type", "application/json")]
            start_response(status, response_headers)
            return [json.dumps({"error": status}).encode("UTF-8")]

    def abort(self, code: int, reason: str) -> NoReturn:
        raise AbortError(f"{code} {reason}")



from typing import Any, Literal
from deta_discord_interactions.models import (
    User,
    Channel,
    Message,
)
from deta_discord_interactions.command import Command


def transform_identifier(obj: Any, on_unknown: Literal['raise', 'ignore'] = 'raise') -> str:
    """Transforms a identifier based on a set of rules.
    If it is not recognized, `unknown` defines whenever to `raise` or `ignore`
    `raise`: Raises a `ValueError` Exception
    `ignore`: Returns as-is
    """
    # Discord models -> their identifiers
    if isinstance(obj, User):  # Also applies for Member
        return f'_discord_interactions_user_{obj.id}'
    elif isinstance(obj, Channel):
        return f'_discord_interactions_channel_{obj.id}'
    elif isinstance(obj, Message):
        if obj.id is None:
            raise Exception("Cannot use a Message without an ID")
        return f'_discord_interactions_message_{obj.id}'
    # Commands -> their name.
    # Not sure if I should support that one, but for now I will
    elif isinstance(obj, Command):
        return f'_discord_interactions_command_{obj.name}'
    # Integers and Strings can stay as-is
    elif isinstance(obj, (int, str)):
        return obj
    # Anything else ERROR
    else:
        if on_unknown == 'raise':
            raise Exception(f"Unsupported object: {obj!r}, type {type(obj)}")
        else:
            return obj



import os
import importlib
import datetime
from typing import Optional
from dataclasses import dataclass

import requests
from deta_discord_interactions.models.user import User
from deta_discord_interactions.models.message import Message

from deta_discord_interactions.models.utils import LoadableDataclass
from deta_discord_interactions.context import Context


@dataclass
class OAuthApplication(LoadableDataclass):
    id: str
    name: str
    icon: str
    description: str
    bot_public: bool
    bot_require_code_grant: bool
    verify_key: str
    hook: bool = False


@dataclass
class OAuthInfo(LoadableDataclass):
    application: OAuthApplication
    scopes: list[str]
    expires: datetime.datetime
    user: User

    def __post_init__(self):
        if isinstance(self.application, dict):
            self.application = OAuthApplication.from_dict(self.application)
        if isinstance(self.user, dict):
            self.user = User.from_dict(self.user)
        if isinstance(self.expires, str):
            self.expires = datetime.datetime.fromisoformat(self.expires)


@dataclass
class OAuthToken(LoadableDataclass):
    access_token: str
    refresh_token: str
    scope: str
    expires_in: int
    webhook: 'Webhook' = None
    expire_date: datetime.datetime = None

    def __post_init__(self):
        if isinstance(self.expires_in, int) and self.expire_date is None:
            self.expire_date = datetime.datetime.utcnow() + datetime.timedelta(seconds=self.expires_in)
        if isinstance(self.webhook, dict):
            self.webhook = Webhook.from_dict(self.webhook)

    def get_user_data(self) -> OAuthInfo:
        headers = {
            "Authorization": f"Bearer {self.access_token}"
        }
        response = requests.get(f'https://discord.com/api/v10/oauth2/@me', headers=headers)
        response.raise_for_status()
        return OAuthInfo.from_dict(response.json())

    def revoke(self) -> None:
        response = requests.post(
            "https://discord.com/api/oauth2/token/revoke",
            data={"token": self.access_token},
            auth=(os.getenv("DISCORD_CLIENT_ID"), os.getenv("DISCORD_CLIENT_SECRET"))
        )
        response.raise_for_status()



@dataclass
class PendingOAuth(LoadableDataclass):
    "A 'promise' of an OAuth process yet to be confirmed"
    ctx: Context
    callback_module: str
    callback_name: str
    callback_args: list
    callback_kwargs: dict

    @property
    def callback(self):
        module = importlib.import_module(self.callback_module)
        function = getattr(module, self.callback_name)
        return function

    def execute_callback(self, oauth_token: OAuthToken):
        "Executes the callback when the user finishes the OAuth process"
        return self.callback(oauth_token, self.ctx, *self.callback_args, **self.callback_kwargs)

    def to_dict(self):
        _discord_interactions = self.ctx.discord
        try:
            self.ctx.discord = None
            return super().to_dict()
        finally:
            self.ctx.discord = _discord_interactions


@dataclass
class Webhook(LoadableDataclass):
    "Represents an active Discord Webhook"
    id: str
    token: str
    name: str
    avatar: str
    guild_id: str = None
    channel_id: str = None
    # Other fields that we do not store:
    # user: User, does not seems to appear in the data?
    # application_id: str, should always match os.getenv('DISCORD_CLIENT_ID')
    # access_token: str, not required
    # refresh_token: str, not required
    # type: int, always 1 for the ones we support

    @property
    def url(self):
        if os.getenv("DONT_REGISTER_WITH_DISCORD", False):
            raise Exception("Cannot interact with Webhook with `DONT_REGISTER_WITH_DISCORD` env. variable set")

        return f"https://discord.com/api/webhooks/{self.id}/{self.token}"

    def send(
        self,
        message: Message,
        *,
        wait_for_response: bool = False,
        username: str = None,
        avatar_url: str = None,
    ) -> Optional[Message]:
        """Sends a Message through this Webhook.
        Parameters
        ----------
        message: Message | str
            Message to send, converted by `Message.from_return_value` before being sent.
        wait_for_response: bool
            Indicates for Discord to send and return a Message object
        username: str
            Overwrites the default username used for this webhook
        avatar_url: str
            Overwrites the default avatar used for this webhook
        
        Returns
        -------
        Message | None
            If `wait_for_response` is set to True, returns the Message.
            Otherwise returns None
        """
        message = Message.from_return_value(message)

        encode_kwargs = {}
        if username is not None:
            encode_kwargs["username"] = username 
        if avatar_url is not None:
            encode_kwargs["avatar_url"] = avatar_url 

        wait_param = 'true' if wait_for_response else 'false'

        encoded, mimetype = message.encode(followup=True, **encode_kwargs)

        response = requests.post(
            self.url,
            data=encoded.encode("UTF-8"),
            headers={"Content-Type": mimetype},
            params={"wait": wait_param}
        )
        response.raise_for_status()
        if wait_for_response:
            return Message.from_dict(response.json())

    def get(self) -> 'Webhook':
        "Returns the updated Discord data for this Webhook"
        response = requests.get(self.url)
        response.raise_for_status()
        return Webhook.from_dict(response.json())

    def patch(self, *, name: Optional[str] = None, avatar: Optional[str] = None, reason: Optional[str] = None):
        """Updates this webhook

        Parameters
        ----------
        name : str
            Change the default name of the Webhook 
        avatar
            NOT SUPPORTED YET
        reason : str
            Audit Log reason explaining why it was updated

        NOTE: Does not updates any internal databases automatically
        """
        if (name is None) and (avatar is None):
            raise ValueError("You must provide at least one of `name` and `avatar` to Webhook.patch()")
        if avatar is not None:
            raise NotImplementedError("Updating the Avatar is not supported by the library yet")
        data = {}
        headers = {}
        if name is not None:
            data["name"] = name
        if reason is not None:
            headers["X-Audit-Log-Reason"] = reason
        response = requests.patch(
            self.url,
            headers=headers,
            json=data,
        )
        response.raise_for_status()
        return response
    
    def delete(self, *, reason: str = None) -> None:
        """Deletes this Webhook
        
        Parameters
        ----------
        reason : str
            Audit Log reason explaining why it was deleted

        NOTE: Does not removes from any internal databases automatically
        """
        headers = {}
        if reason is not None:
            headers["X-Audit-Log-Reason"] = reason
        response = requests.delete(
            self.url,
            headers=headers,
        )
        response.raise_for_status()
    
    def message_url(self, message_id: str) -> str:
        return f'{self.url}/messages/{message_id}'

    def get_message(self, message_id: str) -> Message:
        "Returns a message previously sent through this Webhook"
        response = requests.get(
            self.message_url(message_id),
        )
        response.raise_for_status()
        return Message.from_dict(response.json())

    def delete_message(self, message_id: str) -> None:
        "Deletes a message previously sent through this Webhook"
        response = requests.delete(
            self.message_url(message_id),
        )
        response.raise_for_status()

    def edit_message(self, message: Message, *, message_id: str = None) -> Message:
        """Edits and returns a message previously sent through this Webhook
        
        Parameters
        ----------
        message : Message
            The message to edit it to
        message_id : str
            If present, use this ID instead of message.id

        Must provide a Message with an ID or a message_id
        """
        message_id = message_id or message.id
        if message_id is None:
            raise ValueError("You must provide a message with an ID or a message_id")

        encoded, _ = message.encode(followup=True)

        response = requests.patch(
            self.message_url(message_id),
            data=encoded,
        )
        response.raise_for_status()
        return Message.from_dict(response.json())



import re
import enum
import inspect
import itertools

from typing import Callable, List, Dict, TYPE_CHECKING

from deta_discord_interactions.context import Context
from deta_discord_interactions.models import (
    Message,
    Modal,
    CommandOptionType,
    ApplicationCommandType,
    User,
    Member,
    Channel,
    Role,
    Autocomplete,
    Option,
    Attachment,
)
from deta_discord_interactions.models.autocomplete import AutocompleteResult

if TYPE_CHECKING:
    from deta_discord_interactions.discord import DiscordInteractions

_type = type


class Command:
    """
    Represents an Application Command.

    Attributes
    ----------
    command: Callable
        Function to call when the command is invoked.
    name: str
        Name for this command (appears in the Discord client). If omitted,
        infers the name based on the name of the function.
    name_localizations: Dict[str, str]
        Localization dictionary for name field.
    description: str
        Description for this command (appears in the Discord client). If
        omitted, infers the description based on the docstring of the function,
        or sets the description to "No description", if ``ApplicationCommandType``
        is ``CHAT_INPUT``, else set description to ``None``.
    description_localizations: Dict[str, str]
        Localization dictionary for description field.
    options: List[Option]
        Array of options that can be passed to this command. If omitted,
        infers the options based on the function parameters and type
        annotations.
    annotations: Dict[str, str]
        Dictionary of descriptions for each option provided. Use this only if
        you want the options to be inferred from the parameters and type
        annotations. Do not use with ``options``. If omitted, and if
        ``options`` is not provided, option descriptions default to
        "No description".
    type: int
        Type for this command (depend on the action in the Discord client).
        The value is in ``ApplicationCommandType``. If omitted, set the default
        value to ``ApplicationCommandType.CHAT_INPUT``.
    default_member_permissions: int
        A permission integer defining the required permissions a user must have to run the command
    dm_permission: bool
        Indicates whether the command can be used in DMs
    discord: DiscordInteractions
        DiscordInteractionsBlueprint instance which this Command is associated
        with.
    """

    def __init__(
        self,
        command: Callable,
        name: str,
        description: str,
        *,
        options: List[Option],
        annotations: Dict[str, str],
        type: int = ApplicationCommandType.CHAT_INPUT,
        default_member_permissions: int = None,
        dm_permission: bool = None,
        name_localizations: Dict[str, str] = None,
        description_localizations: Dict[str, str] = None,
        discord: "DiscordInteractions" = None,
        autocomplete_handler: Callable = None,
    ):
        self.command = command
        self.name = name
        self.description = description
        self.options = options
        self.annotations = annotations or {}
        self.type = type
        self.default_member_permissions = default_member_permissions
        self.dm_permission = dm_permission
        self.name_localizations = name_localizations
        self.description_localizations = description_localizations
        self.discord = discord
        self.autocomplete_handler = autocomplete_handler
        self.id = None

        if self.name is None:
            self.name = command.__name__

        if not 1 <= len(self.name) <= 32:
            raise ValueError(
                f"Error adding command {self.name}: "
                "Command name must be between 1 and 32 characters."
            )

        if self.type is ApplicationCommandType.CHAT_INPUT:
            if self.description is None:
                self.description = command.__doc__ or "No description"
            if self.name != self.name.lower():
                raise ValueError(
                    f"Error adding command {self.name}: "
                    "Command name must be fully lowercase. "
                    "No UPPERCASE or CamelCase names are allowed."
                )
            if not re.fullmatch(r"^[\w-]{1,32}$", self.name):
                raise ValueError(
                    f"Error adding command {self.name}: "
                    "Command name does not match regex. "
                    "(Perhaps it contains an invalid character?)"
                )
            if not 1 <= len(self.description) <= 100:
                raise ValueError(
                    f"Error adding command {self.name}: "
                    "Command description must be between 1 and 100 characters."
                )
        else:
            self.description = None

        if self.options:
            self.options = [
                (o.dump() if isinstance(o, Option) else o) for o in self.options
            ]

        if self.type is ApplicationCommandType.CHAT_INPUT and self.options is None:
            sig = inspect.signature(self.command)

            self.options = []
            for parameter in itertools.islice(sig.parameters.values(), 1, None):

                annotation = parameter.annotation
                autocomplete = False

                if _type(annotation) == Autocomplete:
                    annotation = annotation.t
                    autocomplete = True

                # Primitive Types
                if annotation == int:
                    ptype = CommandOptionType.INTEGER
                elif annotation == bool:
                    ptype = CommandOptionType.BOOLEAN
                elif annotation == str:
                    ptype = CommandOptionType.STRING
                elif annotation == float:
                    ptype = CommandOptionType.NUMBER

                # Discord Models
                elif annotation in [User, Member]:
                    ptype = CommandOptionType.USER
                elif annotation == Channel:
                    ptype = CommandOptionType.CHANNEL
                elif annotation == Role:
                    ptype = CommandOptionType.ROLE
                elif annotation == Attachment:
                    ptype = CommandOptionType.ATTACHMENT

                # Enums (Used with choices)
                elif issubclass(annotation, enum.IntEnum):
                    ptype = CommandOptionType.INTEGER
                elif issubclass(annotation, enum.Enum):
                    ptype = CommandOptionType.STRING

                else:
                    raise ValueError(f"Invalid type annotation {annotation}")

                option = {
                    "name": parameter.name,
                    "description": self.annotations.get(
                        parameter.name, "No description"
                    ),
                    "type": ptype,
                    "required": (parameter.default == parameter.empty),
                    "autocomplete": autocomplete,
                }

                if issubclass(annotation, enum.Enum):
                    choices = []

                    if issubclass(annotation, enum.IntEnum):
                        value_type = int
                    else:
                        value_type = str

                    for name, member in annotation.__members__.items():
                        choices.append(
                            {"name": name, "value": value_type(member.value)}
                        )

                    option["choices"] = choices

                self.options.append(option)

    def make_context_and_run(
        self, *, discord: "DiscordInteractions", data: dict
    ):
        """
        Creates the :class:`Context` object for an invocation of this
        command, then invokes itself.

        Parameters
        ----------
        discord: DiscordInteractions
            The :class:`DiscordInteractions` object used to receive this
            interaction.
        data: dict
            The incoming interaction data.

        Returns
        -------
        Message
            The response by the command, converted to a Message object.
        """

        context = Context.from_data(discord, data)
        args, kwargs = context.create_args()

        result = self.run(context, *args, **kwargs)

        if isinstance(result, Modal):
            return result
        else:
            return Message.from_return_value(result)

    def run(self, context: Context, *args, **kwargs):
        """
        Invokes the function defining this command.

        Parameters
        ----------
        context: Context
            The :class:`Context` object representing the current state.
        *args
            Any subcommands of the current command being called.
        **kwargs
            Any other options in the current invocation.
        """

        return self.command(context, *args, **kwargs)

    def make_context_and_run_autocomplete(
        self, *, discord: "DiscordInteractions", data: dict
    ):
        """
        Creates the :class:`Context` object for an invocation of this
        command, then invokes it's autocomplete handler.

        Parameters
        ----------
        discord: DiscordInteractions
            The :class:`DiscordInteractions` object used to receive this
            interaction.
        data: dict
            The incoming interaction data.

        Returns
        -------
        AutocompleteResult
            The response by the handler, converted to an AutocompleteResult object.
        """

        context = Context.from_data(discord, data)
        args, kwargs = context.create_autocomplete_args()

        result = self.run_autocomplete(context, *args, **kwargs)

        return AutocompleteResult.from_return_value(result)

    def run_autocomplete(self, context: Context, *args, **kwargs):
        """
        Invokes the function defining this command's autocomplete.

        Parameters
        ----------
        context: Context
            The :class:`Context` object representing the current state.
        *args
            Any subcommands of the current command being called.
        **kwargs
            Any other options in the current invocation.
        """
        if self.autocomplete_handler is None:
            raise ValueError("This command does not have an autocomplete handler registered")

        return self.autocomplete_handler(context, *args, **kwargs)

    def dump(self):
        "Returns this command as a dict for registration with the Discord API."
        data = {
            "type": self.type,
            "name": self.name,
            "description": self.description,
            "options": self.options,
            "name_localizations": self.name_localizations,
            "description_localizations": self.description_localizations,
        }

        if self.default_member_permissions is not None:
            data["default_member_permissions"] = str(self.default_member_permissions)

        if self.dm_permission is not None:
            data["dm_permission"] = self.dm_permission

        return data

    def autocomplete(self):
        """
        Register an autocomplete handler function for this command.

        Use as a decorator, e.g.

        .. code-block:: python

            @app.command("test")
            def test(ctx, value: str):
                return f"{value}!"

            @test.autocomplete()
            def autocomplete(ctx, value = None):
                return ["hello", "world"]

        """

        def wrapper(function: Callable) -> Callable:
            self.autocomplete_handler = function
            return function

        return wrapper


class SlashCommandSubgroup(Command):
    """
    Represents a Subgroup of slash commands.

    Attributes
    ----------
    name: str
        The name of this subgroup, shown in the Discord client.
    name_localizations: Dict[str, str]
        A dict of localized names for this subgroup.
    description: str
        The description of this subgroup, shown in the Discord client.
    description_localizations: Dict[str, str]
        A dict of localized descriptions for this subgroup.
    """

    def __init__(
        self,
        name: str,
        description: str,
        *,
        name_localizations: Dict[str, str] = None,
        description_localizations: Dict[str, str] = None,
    ):
        self.name = name
        self.description = description
        self.subcommands = {}
        self.type = ApplicationCommandType.CHAT_INPUT
        self.name_localizations = name_localizations
        self.description_localizations = description_localizations

        self.default_member_permissions = None
        self.dm_permission = None

    def command(
        self,
        name: str = None,
        description: str = None,
        *,
        name_localizations: Dict[str, str] = None,
        description_localizations: Dict[str, str] = None,
        options: List[Option] = None,
        annotations: Dict[str, str] = None,
    ):
        """
        Decorator to create a new Subcommand of this Subgroup.

        Parameters
        ----------
        name: str
            The name of the command, as displayed in the Discord client.
        name_localizations: Dict[str, str]
            A dict of localized names for the command.
        description: str
            The description of the command.
        description_localizations: Dict[str, str]
            A dict of localized descriptions for the command.
        options: List[Option]
            A list of options for the command, overriding the function's
            keyword arguments.
        annotations: Dict[str, str]
            If ``options`` is not provided, descriptions for each of the
            options defined in the function's keyword arguments.
        """

        def decorator(func: Callable) -> Command:
            subcommand = Command(
                func,
                name,
                description,
                name_localizations=name_localizations,
                description_localizations=description_localizations,
                options=options,
                annotations=annotations,
            )
            self.subcommands[subcommand.name] = subcommand
            return subcommand

        return decorator

    @property
    def options(self):
        """
        Returns an array of options that can be passed to this command.
        Computed based on the options of each subcommand or subcommand group.

        Returns
        -------
        List[Option]
            The options for this command.
        """
        options = []
        for command in self.subcommands.values():
            data = command.dump()
            if isinstance(command, SlashCommandSubgroup):
                data["type"] = CommandOptionType.SUB_COMMAND_GROUP
            else:
                data["type"] = CommandOptionType.SUB_COMMAND
            options.append(data)

        return options

    def run(self, context, *subcommands, **kwargs):
        """
        Invokes the relevant subcommand for the given :class:`Context`.

        Parameters
        ----------
        context: Context
            The :class:`Context` object representing the current state.
        *args
            List of subcommands of the current command group being invoked.
        **kwargs
            Any other options in the current invocation.
        """
        return self.subcommands[subcommands[0]].run(context, *subcommands[1:], **kwargs)

    def run_autocomplete(self, context, *subcommands, **kwargs):
        """
        Invokes the relevant subcommand's autocomplete handler for the given :class:`Context`.

        Parameters
        ----------
        context: Context
            The :class:`Context` object representing the current state.
        *args
            List of subcommands of the current command group being invoked.
        **kwargs
            Any other options in the current invocation.
        """
        return self.subcommands[subcommands[0]].run_autocomplete(context, *subcommands[1:], **kwargs)


class SlashCommandGroup(SlashCommandSubgroup):
    """
    Represents a Subgroup of slash commands.

    Attributes
    ----------
    name: str
        The name of this subgroup, shown in the Discord client.
    name_localizations: Dict[str, str]
        A dict of localized names for this subgroup.
    description: str
        The description of this subgroup, shown in the Discord client.
    description_localizations: Dict[str, str]
        A dict of localized descriptions for this subgroup.
    default_member_permissions: int
        Permission integer setting permission defaults for a command
    dm_permission: int
        Indicates whether the command can be used in DMs
    """

    def __init__(
        self,
        name,
        description,
        *,
        default_member_permissions=None,
        dm_permission=None,
        name_localizations=None,
        description_localizations=None,
    ):
        self.name = name
        self.description = description
        self.subcommands = {}
        self.type = ApplicationCommandType.CHAT_INPUT

        self.default_member_permissions = default_member_permissions
        self.dm_permission = dm_permission
        self.name_localizations = name_localizations
        self.description_localizations = description_localizations

    def subgroup(
        self,
        name: str,
        description: str = "No description",
        *,
        name_localizations: Dict[str, str] = None,
        description_localizations: Dict[str, str] = None,
    ):
        """
        Create a new :class:`SlashCommandSubroup`
        (which can contain multiple subcommands)

        Parameters
        ----------
        name: str
            The name of the subgroup, as displayed in the Discord client.
        name_localizations: Dict[str, str]
            A dict of localized names for the subgroup.
        description: str
            The description of the subgroup. Defaults to "No description".
        description_localizations: Dict[str, str]
            A dict of localized descriptions for the subgroup.
        """

        group = SlashCommandSubgroup(
            name,
            description,
            name_localizations=name_localizations,
            description_localizations=description_localizations,
        )
        self.subcommands[name] = group
        return group



from __future__ import annotations

import typing
from deta_discord_interactions.utils.database.bound_meta import BoundMeta
if typing.TYPE_CHECKING:
    from deta_discord_interactions.utils.database.record import Record


bind_methods = [
    '__delitem__', '__ior__', '__setitem__', 'clear', 'get', 'pop', 'popitem', 'update',
    'setdefault',
    'get', '__getitem__',
]

class BoundDict(dict, metaclass=BoundMeta, bind_methods=bind_methods):
    """Dictionary which updates the database when modified.
    If you wish to make changes without affecting the database, use dict.copy()
    """
    def __init__(self, bound_key: str, bound_record: 'Record', *argument):
        super().__init__(*argument)
        self._bound_key = bound_key
        self._bound_record = bound_record

    def _sync(self, method, value, *args, **kwargs):
        # 'Getter' methods
        if method in ('get', '__getitem__', 'setdefault'):
            key = kwargs.get('key') if 'key' in kwargs else args[0]
            if isinstance(value, list):
                from deta_discord_interactions.utils.database.bound_list import BoundList
                value = BoundList(
                    f'{self._bound_key}.{key}',
                    self._bound_record,
                    value,
                )
            elif isinstance(value, dict):
                value = BoundDict(
                    f'{self._bound_key}.{key}',
                    self._bound_record,
                    value,
                )

        # Make sure we keep up with inplace methods
        if method.startswith("__i") and isinstance(value, BoundDict):
            self = value

        # 'Setter' methods
        if method not in ('get', '__getitem__'):
            if self._bound_record._preparing_statement:
                self._bound_record._prepared_statement[
                    self._bound_key
                ] = dict(self)
            else:
                self._bound_record._database.update(
                    self._bound_record.key,
                    {self._bound_key: dict(self)}
                )
        return value



class PermissionType:
    ROLE = 1
    USER = 2
    CHANNEL = 3


class Permission:
    """
    An object representing a single permission overwrite.

    ``Permission(role='1234')`` allows users with role ID 1234 to use the
    command

    ``Permission(user='5678')`` allows user ID 5678 to use the command

    ``Permissoin(channel='9012')`` allows users in channel ID 9012 to use the command

    ``Permission(role='3456', allow=False)`` denies users with role ID 3456
    from using the command

    Parameters
    ----------
    role: str
        The role ID to apply the permission to.
    user: str
        The user ID to apply the permission to.
    channel: str
        The channel ID to apply the permission to.
    allow: bool
        Whether use of the command is allowed or denied for the specified criteria.
    """

    def __init__(self, role=None, user=None, channel=None, allow=True):
        if [bool(role), bool(user), bool(channel)].count(True) != 1:
            raise ValueError("specify only one of role, user, or channel")

        if role:
            self.type = PermissionType.ROLE
            self.id = role
        elif user:
            self.type = PermissionType.USER
            self.id = user
        elif channel:
            self.type = PermissionType.CHANNEL
            self.id = channel

        self.permission = allow

    def dump(self):
        """
        Returns a dict representation of the permission.

        Returns
        -------
        dict
            A dict representation of the permission.
        """
        return {"type": self.type, "id": self.id, "permission": self.permission}

    @classmethod
    def from_dict(cls, data: dict):
        """
        Returns a Permission object loaded from a dict.

        Parameters
        ----------
        data: dict
            A dict representation of the permission.
        """
        if data["type"] == PermissionType.ROLE:
            return cls(role=data["id"], allow=data["permission"])
        elif data["type"] == PermissionType.USER:
            return cls(user=data["id"], allow=data["permission"])
        elif data["type"] == PermissionType.CHANNEL:
            return cls(channel=data["id"], allow=data["permission"])
        else:
            raise ValueError("invalid permission type")



"""Query utilities

Example usage:
# Minimal usage
query = Query(Field('name') == 'bob')

query = Query('gaming' in Field('hobbies'))

query = Query(Field('key').startswith('discord_interactions_user_'))

# Combining multiple statements:
query = Query(  # AND
  Field('name') == 'bob',
  Field('age').in_range(10, 18)
)
query = Query([  # AND
  Field('name') == 'bob',
  Field('age').in_range(10, 18)
])
query = Query(  # OR
  [Field('name') == 'bob'],
  [Field('age').in_range(10, 18)]
)
# query = Query(Field('name') == 'bob') | Query(Field('age') > 10)  # Combine as OR CURRENTLY NOT IMPLEMENTED
query = Query(Field('person.name') == 'bob') & Query(Field('age') > 10)  # Combine as AND
query = Query(Field('name') == 'bob') & Query(Field('age') > 10)

# To evaluate a query
results = database.fetch(query)
"""

from typing import NoReturn
from deta_discord_interactions.utils.database.adapters import transform_identifier

class Field:
    "Proxy class for setting filters"
    def __init__(self, attribute: str):
        self.attribute = attribute

    def __eq__(self, other) -> dict:
        other = transform_identifier(other, on_unknown='ignore')
        return {
            self.attribute: other
        }

    def __ne__(self, other) -> dict:
        other = transform_identifier(other, on_unknown='ignore')
        return {
            f'{self.attribute}?ne': other
        }

    def __lt__(self, other) -> dict:
        return {
            f'{self.attribute}?lt': other
        }

    def __gt__(self, other) -> dict:
        return {
            f'{self.attribute}?gt': other
        }

    def __lte__(self, other) -> dict:
        return {
            f'{self.attribute}?lte': other
        }

    def __gte__(self, other) -> dict:
        return {
            f'{self.attribute}?gte': other
        }

    def startswith(self, other) -> dict:
        return {
            f'{self.attribute}?pfx': other
        }
    def prefix(self, other) -> dict:
        return {
            f'{self.attribute}?pfx': other
        }

    def in_range(self, start: int, stop: int) -> dict:
        return {
            f'{self.attribute}?r': [start, stop]
        }

    def __contains__(self, other) -> dict:
        other = transform_identifier(other, on_unknown='ignore')
        return {
            f'{self.attribute}?contains': other
        }
    def contains(self, other) -> dict:
        other = transform_identifier(other, on_unknown='ignore')
        return {
            f'{self.attribute}?contains': other
        }

    def not_contains(self, other) -> dict:
        other = transform_identifier(other, on_unknown='ignore')
        return {
            f'{self.attribute}?not_contains': other
        }


class Query:
    """Query class for querying Deta Base.
    
    Example usage:
    >>> Query(Field("key").startswith("bird_"))  # Key starts with bird_
    >>> Query(Field("species") == "cat", Field("age") < 5)  # Cat AND less than 5 years old
    >>> Query(Field("species") == "cat") | Query(Field("species") == "dog")  # Dog OR Cat
    """
    def __init__(self, *operations: dict):
        self.operations = {}
        for operation in operations:
            if not isinstance(operation, dict):
                raise TypeError("All operations passed to Query() must be dictionaries")
            # Check if there are any operations already in use
            if repeated := (self.operations.keys() & operation.keys()):
                _repeated = {k: (self.operations.get(k), operation.get(k)) for k in repeated}
                raise ValueError(f"Repeated operations found in Query: {_repeated}")
            self.operations.update(operation)

    def __and__(self, other: 'Query') -> 'Query':  # &
        # if isinstance(other, Query):
        if type(other) == Query:
            return Query(self.operations, other.operations)
            # raise NotImplementedError()
        raise TypeError("Can only AND a query with another Query")

    def __or__(self, other: 'Query') -> 'ORQuery':  # |
        # if isinstance(other, Query):
        if type(other) == Query:
            return ORQuery(self, other)
            # raise NotImplementedError()
        raise TypeError("Can only OR a query with another Query")

    def to_list(self) -> list[dict]:
        "Returns a list of length 1 containing this Query's operations"
        return [self.operations]


# tbh it only subclasses so that it may pass isinstance() checks,
# maybe I should be using a Protocol instead
class ORQuery(Query):
    """OR'ed Query class for querying Deta Bases. 
    Do not create directly - use Query(...) | Query(...) instead"""
    def __init__(self, *queries: Query):
        self.queries = queries

    def __and__(self, _) -> NoReturn:  # &
        raise Exception("Cannot add an AND statement to an OR query")

    def __or__(self, other: 'Query') -> 'ORQuery':  # |
        if isinstance(other, ORQuery):
            return ORQuery(*self.queries, *other.queries)
        elif isinstance(other, Query):
            return ORQuery(*self.queries, other)
        raise TypeError("Can only OR with another Query")

    def to_list(self) -> list[list[dict]]:
        "Returns a 1D list containing this queries's operations"
        return [query.operations for query in self.queries]



import functools

class BoundMeta(type):
    """A class made to ease syncing changes to mutable objects to the database.

    When any of the methods called to `bind_methods` are called, it will:
        - execute that method
        - call self._sync() with:
            the name of that method,
            the value returned by that method
            the arguments and keyword arguments passed to that method
        - return the value returned by _sync
    Note: DO NOT overwrite these methods on your class

    Example Usage:
        class Test(list, metaclass=BoundMeta, bind_methods=['__iadd__', 'append', 'extend']):
            def _sync(self, method, value, *args, **kwargs):
                print(self, method, value)
                return value
    """

    def __new__(cls, name, bases, namespace, *, bind_methods):
        if '_sync' not in namespace:
            raise ValueError("Classes using `BoundMeta` as a metaclass must define a `_sync` method")

        new_class = super().__new__(cls, name, bases, namespace)

        for field in bind_methods:

            @functools.wraps(getattr(new_class, field))
            def wrapped(self, *args, _field=field, **kwargs):
                value = getattr(super(new_class, self), _field)(*args, **kwargs)
                return self._sync(_field, value, *args, **kwargs)

            setattr(new_class, field, wrapped)

        return new_class



from deta_discord_interactions.utils.database.database import Database
from deta_discord_interactions.utils.database.record import Record
from deta_discord_interactions.utils.database.query import Query, Field
from deta_discord_interactions.utils.database.bound_dict import BoundDict
from deta_discord_interactions.utils.database.bound_list import BoundList

__all__ = [
    "Database",
    "Record",
    "Query",
    "Field",
    "BoundDict",
    "BoundList"
]


import json
from typing import Union
from dataclasses import dataclass

from deta_discord_interactions.models.message import ResponseType
from deta_discord_interactions.models.option import Choice

class _MetaAutocomplete(type):
    def __getitem__(self, t: type):
        return self(t)


class Autocomplete(metaclass=_MetaAutocomplete):
    """
    Represents the type of an option that can be autocompleted.

    Parameters
    ----------
    t
        The underlying type of the option.
    """

    def __init__(self, t: type):
        self.t = t
    
    # "Using __class_getitem__() on any class for purposes other than type hinting is discouraged."
    # def __class_get_item__(cls, t: type) -> 'Autocomplete':
    #     return cls(t)


@dataclass
class AutocompleteResult:
    """
    Represents the result of an autocomplete handler.

    Parameters
    ----------
    choices
        A list of dicts representing the choices to be presented to the user.
    """
    choices: list[dict]

    def encode(self):
        """
        Return this result as a complete interaction response.

        Returns
        -------
        str
            The encoded JSON object.
        str
            The mimetype of the response (``application/json``).
        """
        data = {
            "type": ResponseType.APPLICATION_COMMAND_AUTOCOMPLETE_RESULT,
            "data": {"choices": self.choices},
        }

        return json.dumps(data), "application/json"

    @classmethod
    def from_return_value(cls, value: Union[dict, list, "AutocompleteResult"]):
        """
        Converts the return value of an autocomplete handler to an
        AutocompleteResult.

        Parameters
        ----------
        value
            The return value of an autocomplete handler.

        Returns
        -------
        AutocompleteResult
            The AutocompleteResult that corresponds to the return value.
        """
        if isinstance(value, AutocompleteResult):
            return value
        elif isinstance(value, dict):
            return cls([value])
        elif isinstance(value, list) and all(
            isinstance(choice, dict) for choice in value
        ):
            return cls(value)
        elif isinstance(value, list) and all(
            isinstance(choice, Choice) for choice in value
        ):
            return cls([choice.dump() for choice in value])
        elif isinstance(value, list):
            return cls(
                [{"name": str(choice), "value": choice} for choice in value]
            )



from __future__ import annotations

import os
import importlib
import itertools
from typing import Callable, Optional, Type, Union
from typing import Generic
from datetime import datetime
import inspect

from deta_discord_interactions.utils.database.exceptions import KeyNotFound
from deta_discord_interactions.utils.database.record import (
    Record,
    AutoSyncRecord,
    RecordType,
    Key,
)
from deta_discord_interactions.utils.database.adapters import transform_identifier
from deta_discord_interactions.utils.database.query import Query

from deta_discord_interactions.models.utils import LoadableDataclass

from deta import Base
from deta_discord_interactions.utils.database._local_base import Base as LocalBase

# Instructions for encoding / decoding data not supported by deta base
EMPTY_LIST_STRING = "$EMPTY_LIST"  # Setting a field to an empty list sets it to `null`
EMPTY_DICTIONARY_STRING = "$EMPTY_DICT"  # Setting a field to an empty dictionaries seems to set it to `null`
DATETIME_STRING = "$ENCODED_DATETIME"  # Ease datetime conversion
ESCAPE_STRING = "$NOOP"  # Do not mess up if the user input 'just happen' to start with a $COMMAND


class Database(Generic[Key, RecordType]):
    def __init__(self, name: str = "_discord_interactions", record_type: Type[RecordType] = AutoSyncRecord):
        """Deta Base wrapper | ORM
        
        Parameters
        ----------
        name : str
            Which name to use for the Deta Base
        record_type : RecordType, default Record
            Which type of Record to return.
            You may subclass the Record class
            if you want to specify which fields each record in that Database should have
        """
        base_mode = os.getenv("DISCORD_INTERACTIONS_DATABASE_MODE", "DETA_BASE")
        if base_mode == "DETA_BASE":
            self.__base = Base(name)
        elif base_mode == "MEMORY":
            self.__base = LocalBase(name)
        elif base_mode == "DISK":
            self.__base = LocalBase(name, sync_disk=True)
        else:
            raise Exception("Invalid value for DISCORD_INTERACTIONS_DATABASE_MODE environment variable")
        self.__record_type = record_type
        self.load_record = record_type.from_database

    def __getitem__(self, key: Key) -> RecordType:
        key = transform_identifier(key)
        return self.load_record(key, self, None)

    def __setitem__(self, key: Key, record: Record) -> None:
        if not isinstance(record, Record):
            raise TypeError("You can only set database values to Record subclasses.")
        self.put(key, record.to_dict())
    
    def __delitem__(self, key: Key) -> None:
        self.delete(key)

    def encode_entry(self, record: Union[dict, Record, LoadableDataclass]) -> dict:
        "Converts values so that we can store it properly in Deta Base. Does not modifies in-place."
        if isinstance(record, (Record, LoadableDataclass)):
            return self.encode_entry(record.to_dict())
        record = record.copy()
        for key, value in record.items():
            if isinstance(value, dict) and dict(value) == {}:  # Empty dict becomes `null` on deta base on update
                record[key] = EMPTY_DICTIONARY_STRING
            elif isinstance(value, list) and list(value) == []:  # Empty lists becomes `null` on deta base on update
                record[key] = EMPTY_LIST_STRING
            elif inspect.isfunction(value):  # Converts functions to references based on their source file and name
                # This should only be used if this record is only going to be stored for a short amount of time
                # And even then, it should be using sparingly
                record[key] = {
                    "__database_load_method": "function",
                    "__module": value.__module__,
                    "__name": value.__name__,
                }
            elif isinstance(value, (dict, Record, LoadableDataclass)):  # Convert nested fields 
                record[key] = self.encode_entry(value)
            elif isinstance(value, list):  # Convert all list elements
                # NOTE: Currently won't work for 2D lists
                record[key] = [
                    self.encode_entry(element) 
                    if isinstance(element, (dict, LoadableDataclass)) else element 
                    for element in value
                ]
            elif isinstance(value, datetime):  # Ease datetime conversion
                record[key] = DATETIME_STRING + value.isoformat()
            elif isinstance(value, str) and value.startswith("$"):  # essentially escape '$'
                record[key] = ESCAPE_STRING + value
        return record

    def _load_encoded(self, record: dict) -> Union[LoadableDataclass, Callable]:
        "Tries to load back an encoded record field into a user defined class or function. May raise exceptions."
        method = record['__database_load_method']
        if method == "function":
            module = record['__module']
            name = record['__name']
            return getattr(importlib.import_module(module), name)
        elif method == "LoadableDataclass.from_dict":
            cls = LoadableDataclass._known_database_models[record["__class_name"]]
            return cls.from_dict(record)

    def decode_entry(self, record: dict) -> Union[dict, LoadableDataclass]:
        "Converts back some changes that we may make when storing. May modify in-place."
        for key, value in record.items():
            if isinstance(value, dict):  # Make sure we hit nested fields
                record[key] = self.decode_entry(value)
            elif isinstance(value, list):  # Convert all list elements. NOTE: Currently won't work for 2D lists
                record[key] = [
                    self.decode_entry(element) if isinstance(element, dict) else element
                    for element in value
                ]
            elif isinstance(value, str):  # Revert our custom 'special' strings
                if value == EMPTY_DICTIONARY_STRING:  # Empty dict becomes `null` on deta base
                    record[key] = {}
                elif value == EMPTY_LIST_STRING:  # Empty lists becomes `null` on deta base
                    record[key] = []
                elif value.startswith(DATETIME_STRING):  # Ease datetime conversion
                    record[key] = datetime.fromisoformat(value.removeprefix(DATETIME_STRING))
                elif value.startswith(ESCAPE_STRING):  # Escape strings starting with `$`
                    record[key] = value.removeprefix(ESCAPE_STRING)

        try:  # Check if it is an encoded Dataclass or function
            return self._load_encoded(record)
        except KeyError:
            pass
        return record

    def get(self, key: str) -> RecordType:
        """Retrieve a record based on it's key. 
        If it does not exists, prepare a blank one with that key"""
        key = transform_identifier(key)
        data = self.__base.get(key)
        if data is None:
            data = {}
        return self.load_record(key, self, self.decode_entry(data))

    def insert(self, key: str, data: dict) -> RecordType:
        "Insert a record and return it."
        key = transform_identifier(key)
        self.__base.insert(self.encode_entry(data), key)
        return self.load_record(key, self, data)
    
    def put(self, key: str, data: dict) -> RecordType:
        "Insert or update a record and return it."
        key = transform_identifier(key)
        self.__base.put(self.encode_entry(data), key)
        return self.load_record(key, self, data)
    
    def delete(self, key: str) -> None:
        "Deletes a record."
        key = transform_identifier(key)
        self.__base.delete(key)

    def _put_many_list(self, data: list[list[Union[dict, Record]]], key_source: Callable[[Union[dict, Record]], str]):
        _all_records = []
        for sublist in data:
            records = []
            for record in sublist:
                _key = key_source(record)
                if isinstance(record, Record):
                    record = record.to_dict()
                record = self.encode_entry(record)
                record['key'] = _key
                records.append(record)
            self.__base.put_many(records)
            _all_records.extend(records)
        return _all_records
    
    def _put_many_dict(self, data: list[tuple[str, Union[dict, Record]]]):
        _all_records = []
        for subdict in data:
            records = []
            for k, record in subdict:
                if isinstance(record, Record):
                    record = record.to_dict()
                record = self.encode_entry(record)
                record['key'] = k
                records.append(record)
            self.__base.put_many(records)
            _all_records.extend(records)
        return _all_records


    def put_many(self,
        data: Union[list[Union[Record, dict]], dict],
        *,
        key_source: Union[str, Callable[[Union[Record, dict]], str]] = 'key',
        iter: bool = False,
    ) -> list[RecordType]:
        """Insert or overwrite multiple records and return them. 
        Deta Base has a limit of up to 25 records at once without `iter`

        Parameters
        ----------
        data : list of records or dictionary
            If a list of records: All records them must have `key_field` set as an attribute or item
            If a dictionary: Uses the dictionary's key for each record
        key_source: str or function
            Which field to use as the `key` for each record in data.
            If it's callable, it will be called for each record
            Ignored when using a dictionary
        iter : bool, default False
            Automatically split the data into sublists of up to 25 items and put multiple times.
        """
        if isinstance(data, list):
            if iter:
                data_chain = itertools.chain(data[offset:offset+25] for offset in range(0, len(data), 25))
            else:
                data_chain = [data]
            if isinstance(key_source, str):
                key_field = key_source
                key_source = lambda record: getattr(record, key_field, None) or record[key_field]
            _all_records = self._put_many_list(data_chain, key_source)
        elif isinstance(data, dict):
            if iter:
                _it = iter(data.items())
                data_chain = (itertools.islice(_it, 25) for _ in range(((len(data)-1) // 25) + 1))
            else:
                data_chain = [data.items()]
            _all_records = self._put_many_dict(data_chain)
        else:
            raise TypeError(f"Unsupported type {type(data)} passed to database.put_many: {data!r}")
        return [self.load_record(record["key"], self, self.decode_entry(record)) for record in _all_records]

    def fetch(
        self,
        query: Union[Query, dict, list[dict], None] = None,
        limit: int = 1000,
        last: Optional[str] = None,
        follow_last: bool = False,
    ) -> list[RecordType]:
        """Returns multiple items from the database based on a query.

        Parameters
        ----------
        query : Query
            See the `Query` and `Field` classes as well as https://docs.deta.sh/docs/base/queries/ for more information.
        limit : int, default 1000
            Maximum number of records to fetch.
            NOTE: Deta Base will only retrieve up to 1MB of data at a time, and that is before applying the filters
        last : str, default None
            Equivalent to `offset` in normal databases, but key based instead of position based
        follow_last : bool, default False
            Automatically fetch more records up to `limit` if the query returns a `last` element
        """
        if isinstance(query, Query):
            query = query.to_list()
        result = self.__base.fetch(query, limit=limit, last=last)
        records = result.items
        if follow_last:
            while result.last is not None and len(records) < limit:
                result = self.__base.fetch(query, limit=limit, last=result.last)
                records.extend(result.items)

        return [
            self.load_record(record['key'], self, self.decode_entry(record))
            for record in records
        ]

    def update(self, key: str, updates: dict) -> RecordType:
        """Updates a Record.

        Note: The returned Record may have to fetch back the updated data
        """
        key = transform_identifier(key)

        updates = self.encode_entry(updates)

        try:
            self.__base.update(updates, key)
        except Exception as err:
            import re
            reason = err.args[0] if err.args else ''
            if isinstance(reason, str) and re.fullmatch(r"Key \'.*\' not found", reason):
                raise KeyNotFound(reason)
            else:
                raise
        return self.load_record(key, self, None)



"Deals with OAuth2, creating and saving Webhooks"
import json
import os
from typing import Callable, NoReturn
from urllib.parse import quote, unquote

import requests
from deta_discord_interactions.models.component import ActionRow, Button, ButtonStyles

from deta_discord_interactions.models.message import Message

from deta_discord_interactions.discord import DiscordInteractions
from deta_discord_interactions.context import Context

from deta_discord_interactions.utils.database import Database
from deta_discord_interactions.utils.oauth.model import OAuthToken, PendingOAuth


DISCORD_BASE_URL = 'https://discord.com/api/v10'
DEFAULT_MICRO_PATH = "https://{MICRO}.deta.dev"

pending_oauths = Database(name="_discord_interactions_pending_oauths")


def enable_oauth(app: DiscordInteractions, /, *, path: str = "/oauth") -> None:
    "Allows for the app to receive and process OAuth and create Webhooks"
    app.route(path)(_handle_oauth)


def request_oauth(
    ctx: Context,
    /,
    internal_id: str,
    *,
    domain: str = DEFAULT_MICRO_PATH,
    path: str = "/oauth",
    scope: str,
    callback: Callable,
    args: tuple = [],
    kwargs: dict = {},
    message_content: str = "Use the button to register with OAuth",
    button_label: str = "Grant OAuth",
) -> Message:
    """Utility function to make OAuth creation and usage easier
    
    Returns a Message with a link the user must visit to grant an OAuth Token,
    and save a PendingOAuth in the internal database.

    See https://discord.com/developers/docs/topics/oauth2 for the list of available scopes and more information

    Parameters
    ----------
    ctx : Context
        The Context this function is being called from
    internal_id : str
        ID to be used internally. Will be shown in the link.. Will be shown in the link.
    domain : str, default https://{MICRO}.deta.dev
        Base URL for the Micro running the bot
        {MICRO} is filled automatically from the environment variables
    path : str, default '/oauth'
        Path that the user will be sent back to. 
        Must match what has been passed to `enable_webhooks` and be set on the Developer Portal
    scope : str
        OAuth scopes to request, separated by spaces.
    callback : Callable
        Must be a normal function, not a lambda, partial nor a class method.
    args : tuple|list
    kwargs : dict
        Arguments and Keyword arguments to be passed onto callback
        The created newly created OAuthToken and Context passed to request_oauth are always passed first.
    message_content : str
        Message content, besides the button with the URL
    button_label : str
        The URL Button's label


    If the user never finishes the authorization process, the callback will not be called
    If they create one , it will be called with ctx, webhook, `args` and `kwargs`
    The link will only work for one authorization
    """
    promise = PendingOAuth(
        ctx,
        callback.__module__,
        callback.__name__,
        args,
        kwargs,
    )
    redirect_uri = (
        quote(
            domain.format(MICRO = os.getenv("DETA_PATH")) + path,
            safe=''
        )
    )

    with pending_oauths[internal_id] as record:
        record["pending_oauth"] = promise
        record["redirect_uri"] = redirect_uri

    link = (
        f"{DISCORD_BASE_URL}/oauth2/authorize?"
        "response_type=code&"
        f"scope={quote(scope)}&"
        f"guild_id={ctx.guild_id}&"
        f"client_id={os.getenv('DISCORD_CLIENT_ID')}&"
        f"state={internal_id}&"
        f"redirect_uri={redirect_uri}"
    )

    return Message(
        message_content,
        components=[ActionRow([
            Button(
                style=ButtonStyles.LINK,
                url=link,
                label=button_label,
            )
        ])],
        ephemeral=True,
    )


# Mostly for convenience and more... idk, semantic
def create_webhook(
    ctx: Context,
    /,
    internal_id: str,
    domain: str = DEFAULT_MICRO_PATH,
    path: str = "/oauth",
    *,
    callback: Callable,
    args: tuple = (),
    kwargs: dict = {},
    message_content: str = "Use the button to register the Webhook",
    button_label: str = "Create Webhook",
) -> Message:
    """Utility function to make Webhook creation and usage easier
    
    Returns a Message with a link the user must visit to create a webhook,
    and save a PendingWebhook in the internal database.

    Parameters
    ----------
    ctx : Context
        The Context this function is being called from
    internal_id : str
        ID to be used internally. Will be shown in the link.
    domain : str, default https://{MICRO}.deta.dev
-        Base URL for the Micro running the bot
        {MICRO} is filled automatically from the environment variables
    path : str, default '/oauth'
        Path that the user will be sent back to. 
        Must match what has been passed to `enable_webhooks` and be set on the Developer Portal
    callback : Callable
        Must be a normal function, not a lambda, partial nor a class method.
    args : tuple|list
    kwargs : dict
        Arguments and Keyword arguments to be passed onto callback.
        The created newly created OAuthToken and Context passed to create_webhook are always passed first.
    message_content : str
        Message content, besides the button with the URL
    button_label : str
        The Label of the button


    If the user never finishes creating a webhook, the callback will not be called
    If they create one , it will be called with ctx, webhook, `args` and `kwargs`
    The link will only work for one webhook
    """
    return request_oauth(
        ctx,
        internal_id,
        domain=domain,
        path=path,
        scope='webhook.incoming',
        callback=callback,
        args=args,
        kwargs=kwargs,
        message_content=message_content,
        button_label=button_label,
    )



def _handle_oauth(
    request: dict,
    start_response: Callable[[str, list], None],
    abort: Callable[[int, str], NoReturn],
) -> list[str]:
    try:
        code = request["query_dict"]["code"]
        state = request["query_dict"]["state"]
        # guild_id = request["query_dict"]["guild_id"]
    except KeyError:
        abort(400, 'Invalid URL')

    url = DISCORD_BASE_URL + "/oauth2/token"

    try:

        with pending_oauths[state] as record:
            redirect_uri: str = record["redirect_uri"]
            pending_oauth: PendingOAuth = record["pending_oauth"]

        del pending_oauths[state]

        data = {
            'client_id': os.getenv("DISCORD_CLIENT_ID"),
            'client_secret': os.getenv("DISCORD_CLIENT_SECRET"),
            'grant_type': 'authorization_code',
            'code': code,
            'redirect_uri': unquote(redirect_uri),
        }
        headers = {
            'Content-Type': 'application/x-www-form-urlencoded'
        }
        response = requests.post(url, data=data, headers=headers)
        response.raise_for_status()
        result = response.json()

        oauth_token = OAuthToken.from_dict(result)

        callback_response = pending_oauth.execute_callback(oauth_token)

        if isinstance(callback_response, (dict, list, int, str)):
            callback_response = json.dumps(callback_response)
        if isinstance(callback_response, str):
            callback_response = callback_response.encode("UTF-8")
        if not isinstance(callback_response, bytes):
            raise Exception("The Callback response should be a dictionary, a string or bytes")


        start_response("200 OK", [('Content-Type', 'application/json')])
        return [callback_response]

    except Exception:
        import traceback
        traceback.print_exc()
        try:
            print('response content', response.content, flush=True)
        except Exception:
            pass
        abort(500, "Something went wrong")



from __future__ import annotations

import functools
import json
import pathlib
from typing import Callable, Optional

from deta.base import Util

from deta_discord_interactions.utils.database.bound_meta import BoundMeta

# In the future: Perhaps get rid of this and use pytest fixtures instead?

operations = {
    # no prefix = eq
    "ne": lambda key, value, record: record.get(key) != value,

    "lt": lambda key, value, record: record.get(key) < value,
    "lte": lambda key, value, record: record.get(key) <= value,
    "gt": lambda key, value, record: record.get(key) > value,
    "gte": lambda key, value, record: record.get(key) >= value,

    "pfx": lambda key, value, record: record.get(key).startswith(value),
    "r": lambda key, value, record: record.get(key) in range(*value),
    "contains": lambda key, value, record: value in record.get(key),
    "not_contains": lambda key, value, record: value not in record.get(key),
}

def parse_filter(filter_: dict) -> Callable[[dict], bool]:
    result = []
    for condition, value in filter_.items():
        if isinstance(value, list):
            raise Exception("Nested queries are not supported for local base yet")

        if "?" in condition:
            key, operation = condition.rsplit("?", 1)
            result.append(functools.partial(operations[operation], key, value))
        else:
            key = condition
            result.append(lambda record, _key=key, _value=value: record.get(_key) == _value)

    def match_record(record: dict) -> bool:
        return all(function(record) for function in result)
    return match_record

def parse_filters(filters: Optional[list[dict]]) -> Callable[[dict], bool]:
    if filters is None:
        return lambda _: True
    parsed_filters = [parse_filter(filter_) for filter_ in filters]

    def match_record(record: dict) -> bool:
        return any(filter_(record) for filter_ in parsed_filters)

    return match_record

class FetchResponse:
    def __init__(self, count=0, last=None, items=[]):
        self.count = count
        self.last = last
        self.items = items


_memory_inventory = {}
class Base:
    def __init__(self, name: str, sync_disk: bool = False):
        self.name = name
        if sync_disk:
            self.inventory = DiskBaseBackend(name)
        else:
            self.inventory = _memory_inventory.setdefault(name, {})

    def get(self, key):
        obj = self.inventory.get(key)
        if obj:
            obj = json.loads(obj)
        return obj

    def insert(self, data, key):
        if key in self.inventory:
            raise Exception(f"Item with key '{key}' already exists")
        self.inventory[key] = json.dumps(data)

    def update(self, updates, key):
        if key not in self.inventory:
            raise Exception(f"Key '{key}' not found")

        obj = json.loads(self.inventory[key])

        for attribute, value in updates.items():
            if isinstance(value, Util.Trim):
                del obj[attribute]
            elif isinstance(value, Util.Increment):
                obj[attribute] += value.val
            elif isinstance(value, Util.Append):  # Despite the name, value.val is always a list here
                obj[attribute].extend(value.val)
            elif isinstance(value, Util.Prepend):
                obj[attribute] = (
                    value.val + obj[attribute]
                )
            else:
                obj[attribute] = value
        
        self.inventory[key] = json.dumps(obj)

    def put_many(self, items):
        items = {record.pop('key'): json.dumps(record) for record in (item.copy() for item in items)}
        self.inventory.update(items)

    def put(self, item, key):
        self.inventory[key] = json.dumps(item)

    def fetch(self, query, limit, last):
        results = []
        match_condition = parse_filters(query)
        for key, value in self.inventory.items():
            obj = json.loads(value)
            obj["key"] = key
            if match_condition(obj):
                results.append(obj)
        if isinstance(last, str):
            index = next((i for i, record in enumerate(results) if record['key'] == last), None)
            results = results[index:]
        if limit and limit > 0:
            results = results[:limit]
        return FetchResponse(items=results)




bind_methods = [
    '__delitem__', '__setitem__', 'clear', 'get', 'pop', 'popitem', 'update', 'setdefault'
]

import pathlib
import os


class DiskBaseBackend(dict, metaclass=BoundMeta, bind_methods=bind_methods):
    """Database backend that saves to disk whenever it's modified."""
    def __init__(self, database_name: str):
        try:
            with (pathlib.Path(os.getenv("DISCORD_INTERACTIONS_DATABASE_FOLDER")) / database_name).open('r') as file:
                super().__init__(json.load(file))
        except Exception:
            super().__init__()
        self._file_name = database_name
        if os.getenv("DISCORD_INTERACTIONS_DATABASE_FORMAT_NICELY", False):
            self._options = {
                "indent": 4,
                "sort_keys": True,
            }
        else:
            self._options = {
                "indent": None,
                "separators": (',', ':'),
            }


    def _sync(self, method, value, *args, **kwargs):
        with (pathlib.Path(os.getenv("DISCORD_INTERACTIONS_DATABASE_FOLDER")) / self._file_name).open('w') as file:
            json.dump(dict(self), file, **self._options)
        return value



import dataclasses

from deta_discord_interactions.models.utils import LoadableDataclass
from deta_discord_interactions.models.user import User


@dataclasses.dataclass
class MessageInteraction(LoadableDataclass):
    """
    Partial data of the interaction that a message is a reply to.

    Attributes
    ----------
    id
        The ID (snowflake) of that Interaction.
    name
        The name of the Interaction's application command, as well as the subcommand and subcommand group, where applicable
    type
        The type of the Interaction.
    user
        The User that invoked the interaction.
"""
    id: str
    name: str
    type: int
    user: User
    
    def __post_init__(self):
        if isinstance(self.user, dict):
            self.user = User.from_dict(self.user)



import inspect
from typing import Optional


class LoadableDataclass:
    _known_database_models = {}
    _database_model_name = None

    def __init_subclass__(cls, database_model_name: Optional[str] = None) -> None:
        "Register known Loadable Dataclasses as models for encoding/decoding database records"
        # Intentionally not using `__qualname__` to make it not care about where it is defined
        name = database_model_name or cls.__name__
        if (existing_model := LoadableDataclass._known_database_models.get(name)) is not None:
            raise Exception(f"Cannot create class {cls} with same database model name as {existing_model=}")
        LoadableDataclass._known_database_models[name] = cls
        cls._database_model_name = name

    @classmethod
    def from_dict(cls, data):
        """
        Construct the dataclass from a dictionary, skipping any keys
        in the dictionary that do not correspond to fields of the class.

        Parameters
        ----------
        data: dict
            A dictionary of fields to set on the dataclass.
        """
        return cls(
            **{k: v for k, v in data.items() if k in inspect.signature(cls).parameters}
        )

    def to_dict(self):
        """
        Construct a dictionary from the dataclass.
        """
        cls = type(self)
        dictionary = {
            # If we ever feel like reducing the amount of data stored: that + make sure all model fields have ...=None
            # k: value for k in inspect.signature(cls).parameters
            # if (value := getattr(self, k, None)) is not None
            k: getattr(self, k, None) for k in inspect.signature(cls).parameters
        }
        for key, value in dictionary.items():
            if isinstance(value, LoadableDataclass):
                dictionary[key] = value.to_dict()
        dictionary["__database_load_method"] = "LoadableDataclass.from_dict"
        dictionary["__class_name"] = cls._database_model_name
        return dictionary



from __future__ import annotations

import typing
from deta.base import Util
from deta_discord_interactions.utils.database.bound_meta import BoundMeta
if typing.TYPE_CHECKING:
    from deta_discord_interactions.utils.database.record import Record


bind_methods = [
    '__add__', '__delitem__', '__iadd__', '__imul__', '__setitem__',
    # 'append',
    'clear', 'extend', 'insert', 'pop', 'remove', 'reverse', 'sort',
]

class BoundList(list, metaclass=BoundMeta, bind_methods=bind_methods):
    """List which updates the database when modified.
    If you wish to make changes without affecting the database, use list.copy()
    """
    def __init__(self, bound_key: str, bound_record: 'Record', *argument):
        super().__init__(*argument)
        self._bound_key = bound_key
        self._bound_record = bound_record

    def append(self, item):
        super().append(item)
        if self._bound_record._preparing_statement:
            self._bound_record._prepared_statement[
                self._bound_key
            ] = list(self)
        else:
            self._bound_record._database.update(
                self._bound_record.key,
                {self._bound_key: Util.Append(item)}
            )

    def _sync(self, method, value, *args, **kwargs):
        if method.startswith("__i") and isinstance(value, BoundList):
            self = value
        if self._bound_record._preparing_statement:
            self._bound_record._prepared_statement[
                self._bound_key
            ] = list(self)
        else:
            self._bound_record._database.update(
                self._bound_record.key,
                {self._bound_key: list(self)}
            )
        return value

    # def __getitem__(self, index):
    #     result = super().__getitem__(index)
    #     if not isinstance(index, int):
    #         return result
    #     if isinstance(result, list):
    #         result = BoundList(
    #             f'{self._bound_key}.{index}',
    #             self._bound_record,
    #             result,
    #         )
    #     elif isinstance(result, dict):
    #         from flask_discord_interactions.utils.database.bound_dict import BoundDict
    #         result = BoundDict(
    #             f'{self._bound_key}.{index}',
    #             self._bound_record,
    #             result,
    #         )
    #     return result



from setuptools import setup, find_packages


with open("README.md") as file:
    readme = file.read()


setup(
    name="Deta-Discord-Interactions",
    version="0.0.2",
    url="https://github.com/etrotta/flask-discord-interactions/tree/deta",
    author="etrotta",
    author_email="etrotta@duck.com",
    description="A web framework for Discord interactions specialized for deta.",
    long_description=readme,
    long_description_content_type="text/markdown",
    packages=find_packages(),
    zip_safe=False,
    include_package_data=True,
    platforms="any",
    install_requires=["requests", "PyNaCl", "requests-toolbelt", "deta"],
    tests_require=["pytest"],
    classifiers=[
        "Environment :: Web Environment",
        "Intended Audience :: Developers",
        "Operating System :: OS Independent",
        "License :: OSI Approved :: MIT License",
        "Programming Language :: Python",
    ],
    python_requires=">=3.9",
)



from deta_discord_interactions.utils.oauth.oauth import (
    enable_oauth,
    request_oauth,
    create_webhook,
)
from deta_discord_interactions.utils.oauth.model import (
    OAuthToken,
    Webhook,
    PendingOAuth,
    OAuthInfo,
    OAuthApplication,
)

__all__ = [
    'enable_oauth',
    'request_oauth',
    'create_webhook',

    "OAuthToken",
    "Webhook",
    "OAuthInfo",
    "OAuthApplication",
    "PendingOAuth",
]


from __future__ import annotations

import typing
import dataclasses
from typing import Any, Optional
from typing import TypeVar
from deta.base import Util
from deta_discord_interactions.utils.database.bound_list import BoundList
from deta_discord_interactions.utils.database.bound_dict import BoundDict
from deta_discord_interactions.utils.database.exceptions import FieldNotFoundError, KeyNotFound
if typing.TYPE_CHECKING:
    from deta_discord_interactions.utils.database.database import Database


class Record:
    """Base class used to interface with the Database.

    Direct usage is not recommended. Either:
    - Use AutoSyncRecord
    - Subclass and use with dataclasses @dataclass
    - Subclass and overwrite `__init__`, (classmethod) `from_database`, `to_dict`
    """
    def __init__(self, **kwargs):
        for k, v in kwargs.items():
            setattr(self, k, v)

    @classmethod
    def from_database(cls, key: str, database: 'Database', record: Optional[dict]) -> 'RecordType':
        if record is None:
            record = {}
        if dataclasses.is_dataclass(cls):
            data = {
                field.name: value for field in dataclasses.fields(cls)
                if (value := record.get(field.name)) is not None
            }
        else:
            data = {k: v for k, v in record if not k.startswith("_")}
        try:
            return cls(**data)
        except TypeError:
            raise TypeError(f"Error while loading class {cls!r}, parameters supplied: {data}")

    def to_dict(self) -> dict:
        if dataclasses.is_dataclass(self):
            data = dataclasses.asdict(self)
        else:
            data = {k: v for k, v in vars(self).items() if not k.startswith("_")}
        return data

# Typing stuff
Key = TypeVar("Key")
RecordType = TypeVar("RecordType", bound=Record)

RAISE_ERROR = object()  # Sentinel

class AutoSyncRecord(Record):
    "Record subclass that syncs database changes automatically"

    def __init__(self, key: Key, database: 'Database', data: Optional[dict]):
        self._key = key
        self._database = database
        self._data = data
        self._preparing_statement = False
        self._prepared_statement = {}

    @classmethod
    def from_database(cls, key: Key, database: 'Database', record: Optional[dict]) -> 'AutoSyncRecord':
        return cls(key, database, record)

    def to_dict(self) -> dict:
        self.sync_fields()
        return self._data.copy()

    def sync_fields(self, *, force: bool = False) -> None:
        if (self._data is None) or force:
            self._data = self._database.get(self._key)._data

    def convert_value(self, field: str, value: Any) -> Any:
        if isinstance(value, list):
            result = BoundList(field, self, value)
        elif isinstance(value, dict):
            result = BoundDict(field, self, value)
        else:
            result = value
        return result

    def set_field(self, field: str, value: Any) -> None:
        if self._preparing_statement:
            self._prepared_statement[field] = value
        else:
            try:
                self._database.update(self._key, {field: value})
            except KeyNotFound:
                self._database.put(self._key, {field: value})
            self._data = None

    def get_field(self, field: str, default: Any = RAISE_ERROR) -> Any:
        if field in self._prepared_statement:
            raise Exception("Cannot retrieve back prepared values before syncing with the database.")
        self.sync_fields()
        try:
            result = self._data[field]
        except KeyError:
            if default is RAISE_ERROR:
                raise FieldNotFoundError
            else:
                result = default
        result = self.convert_value(field, result)
        return result

    def delete_field(self, field: str) -> None:
        if self._preparing_statement:
            self._prepared_statement[field] = Util.Trim()
        else:
            del self._data[field]
            self._database.update(self._key, {field: Util.Trim()})

    def __enter__(self):
        self._preparing_statement = True
        return self

    def __exit__(self, exc_type, exc_value, exc_traceback):
        self._preparing_statement = False
        if exc_type is not None:
            import warnings
            warnings.warn("Aborting prepared operation since an Exception was raised.")
        elif self._prepared_statement:
            try:
                self._database.update(self._key, self._prepared_statement)
            except KeyNotFound:
                self._database.put(self._key, {})
                self._database.update(self._key, self._prepared_statement)
        self._prepared_statement = {}

    def __iter__(self):
        self.sync_fields()
        yield from self._data.items()

    def __getattr__(self, attribute: str) -> Any:
        if attribute.startswith("_"):
            raise Exception("Record fields must not start with `_`")
        try:
            return self.get_field(attribute)
        except KeyError:
            raise AttributeError

    def __setattr__(self, attribute: str, value: Any) -> None:
        if attribute.startswith("_"):
            return super().__setattr__(attribute, value)
        self.set_field(attribute, value)

    def __delattr__(self, attribute: str) -> None:
        self.delete_field(attribute)

    def __getitem__(self, key: str) -> Any:
        return self.get_field(key)

    def __setitem__(self, key: str, value: Any) -> None:
        self.set_field(key, value)

    def __delitem__(self, key: str) -> None:
        self.delete_field(key)

    def keys(self) -> list[str]:
        self.sync_fields()
        return self._data.keys()

    def get(self, key: str) -> Any:
        return self.get_field(key, None)

    def setdefault(self, key: str, value: Any) -> Any:
        try:
            return self.get_field(key)
        except FieldNotFoundError:
            value = self.convert_value(key, value)
            self.set_field(key, value)
            return value
